/*!
 * Opengular.js
 * Copyright (c) 2015, the Opengular.js Authors
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 *  purpose with or without fee is hereby granted, provided that the above
 *  copyright notice and this permission notice appear in all copies.
 */
var Opengular = (function() {
	"use strict";
	
	if (typeof jQuery !== "function") {
		throw new Error("jQuery must be loaded before Opengular.js");
	}
	
	var getBaseName = function(str) {
		var sbpos = str.indexOf("[");
		return (sbpos === -1) ? str : str.substr(0, sbpos);
	};
	
	// -[SCOPE CLASS]----------------
	
	var Scope = function() {
		this.__listeners = []; // Array of [ string Key, callable ] arrays
		this.__parent = null;
		this.__suspend = 0;
	};
	
	Scope.prototype.__getScopeForKey = function(val) { // nested scopes
		if (this.hasOwnProperty(val)) {
			return this;
		} else if (this.__parent !== null) {
			return this.__parent.__getScopeForKey(val);
		} else {
			return this; // unknown key - store it in current scope
		}
	};
	Scope.prototype.__get = function(val) {
		if (val.indexOf("[") === -1) {
			var s = this.__getScopeForKey(val);
			return s[val];
			
		} else {
			// Bound to compound string (e.g. foo[0][baz]).
			var baseName = val.substr(0, val.indexOf("["));
			var parts = val.match(/\[[^\]]+\]/g);
			var tmp = this.__get(baseName);
			for (var i = 0, e = parts.length; i !== e; ++i) {				
				tmp = tmp[ parts[i].slice(1, -1) ];
			}
			return tmp;
		}
	};
	Scope.prototype.__set = function(val, data) {
		if (val.indexOf("[") === -1) {
			var s = this.__getScopeForKey(val);
			s[val] = data;
		
		} else {
			// Bound to compound string (e.g. foo[0][baz]).
			var baseName = val.substr(0, val.indexOf("["));
			var parts = val.match(/\[[^\]]+\]/g);
			var s = this.__getScopeForKey(baseName);
			
			var tmp = s[baseName];
			for (var i = 0, e = parts.length - 1; i !== e; ++i) {				
				tmp = tmp[ parts[i].slice(1, -1) ];
			}
			tmp[ parts[parts.length-1].slice(1, -1) ] = data;
		}
	};
	Scope.prototype.__refresh = function(key) {
		if (this.__suspend > 0) {
			return; // Updates are suspended for this component
		}
		
		key = key ? getBaseName(key) : null;
		
		// We repeat the calling process until values stabilise.
		var before = JSON.stringify(key ? this[key] : this);
		
		for (var i = 0; i < this.__listeners.length; ++i) {
			if (key !== null && this.__listeners[i][0] !== key) {
				// A key was set and this is not a good listener for it. Skip
				continue;
			}
			this.__listeners[i][1]();
		}
		
		var after = JSON.stringify(key ? this[key] : this);
		if (before !== after) {
			return this.__refresh(null); // tail call
		}
		
		// We're stable, propagate (either whole-self, or parent scope)
		if (key !== null) {
			return this.__refresh(null); // first time
			
		} else if (this.__parent !== null) {
			return this.__parent.__refresh(null); // second time
		}
	};
	Scope.prototype.addChangeListener = function(for_key, fn, extra_data) {
		this.__listeners.push([getBaseName(for_key), fn, extra_data]);
	};
	Scope.prototype.clearListeners = function() {
		this.__listeners = [];
	};
	
	// -[REGISTRATION]---------------
	
	var controllers = {};
	var directives = [];
	
	var registerController = function(name, fn) {
		controllers[name] = fn;
	};
	
	var registerDirective = function(attrName, fn) {
		directives.push([attrName, fn]);
	};
	
	// -[BUILTINS]-------------------
	
	registerDirective('og-bind', function($el, scope) {
		var bindTo = $el.attr('og-bind');
		
		if ($el.is('input')) {
			$el.on('change keydown keyup', function() {
				var val = $(this).val();
				if (scope.__get(bindTo) != val) { // Dirty-checking
					scope.__set(bindTo, val);
					scope.__refresh();
				}
			});
			scope.addChangeListener(bindTo, function() {
				$el.val( scope.__get(bindTo) );
			});
			
		} else if ($el.is('div') || $el.is('span')) {
			scope.addChangeListener(bindTo, function() {
				$el.text( scope.__get(bindTo) );
			});
			
		}
	});
	
	registerDirective('og-click', function($el, scope) {
		var attrName = $el.attr('og-click');
		
		// Extract arguments in the form FOO(BAR,BAZ) if present
		var arglist = [];
		if (attrName.indexOf('(') !== -1) {
			arglist = attrName.slice(attrName.indexOf('(')+1, -1).split(','); // FIXME: this needs a proper parser
			attrName = attrName.substr(0, attrName.indexOf('('));
		}
		
		// Apply DOM event handler
		$el.on('click', function(ev) {
			var cb = scope.__get(attrName);
			// Look up arguments from current scope (if any)
			var argmap = [];
			for (var i = 0, e = arglist.length; i !== e; ++i) {
				argmap.push( scope.__get(arglist[i]) );
			}
			var retn = cb(ev, argmap);
			scope.__refresh();
			return retn;
		});
	});
	
	registerDirective('og-repeat', function($el, scope) {
		var matches = $el.attr('og-repeat').match(/([^ ]+) in ([^ ]+)/); // TODO: support custom names for $index variable
		if (matches === null) return; // invalid
		
		var arrName = matches[2];
		var subName = matches[1];
		
		var stashChildCount = $el.children().length;
		var stashHtml = $el.html();
		
		var rebuild = function() {
			// Since this calls initDiv() which calls __refresh() which calls the rebuild() change listener, we must apply a lock to prevent infinite loop
			scope.__suspend++;
			
			// Remove old scopes and old HTML as best as we can
			
			$el.find("[og-managed]").each(function() {
				$(this).data('og-scope').clearListeners();
			});
			$el.html(''); // wiping any event handlers that might be present
			scope.__listeners = scope.__listeners.filter(function(el) { // remove old references from parent to child
				return (el[2] /* extra_data */ !== "og-repeat::"+arrName);
			});
			
			for (var i = 0, e = scope.__get(arrName).length; i !== e; ++i) {
				var $child = $(stashHtml);
				$el.append($child);
				
				var makeChildScopeFor = function(scope, arrName, i, subName) {
					var childScope = new Scope();
					childScope.__parent = scope;
					
					var containingScope = scope.__getScopeForKey(arrName);
					childScope[subName] = containingScope[arrName][i]; // equivalent to scope.__get(arrName)[i], but containingScope is reused
					childScope["$index"] = i; // create a new special value
					childScope["$length"] = containingScope[arrName].length;
					
					childScope.addChangeListener(subName, function() {
						if (childScope.__parent.__get(arrName).length != childScope["$length"]) {
							return; // In the process of tearing down/rebuilding, don't set
						}
						childScope.__parent.__get(arrName)[i] = childScope[subName];
						// parent refresh is called automatically...
					});
					
					containingScope.addChangeListener(arrName, function() {
						// ...but child refresh is not called automatically!
						// We have to manually call the refresh (and perform our own dirty checking to prevent stack overflow)
						if (childScope[subName] != containingScope[arrName][i]) {
							childScope[subName] = containingScope[arrName][i];					
							childScope.__refresh(subName);
						}
					}, "og-repeat::"+arrName );
					
					
					return childScope;
				};
				
				var childScope = makeChildScopeFor(scope, arrName, i, subName);
				$child.data('og-scope', childScope);
				initDiv( $child, childScope );
			};
			
			scope.__suspend--;
			
			scope.__refresh();
		};
		
		scope.addChangeListener(arrName, function() {
			// Only call rebuild() if the number of children changed, since internal element changes should be handled by individual og-bind directives.
			var expectedCount = scope.__get(arrName).length;
			var currentCount = $el.children().length / stashChildCount;
			if (expectedCount != currentCount) {
				rebuild();
			}
		});
		
		rebuild();
	});
	
	// -[STARTUP]--------------------
	
	var initDiv = function($el, scope) {
		// Don't allow double-registration
		if ($el.attr('og-managed') === "") return;
		$el.attr('og-managed', '');
		
		// Register directives
		for (var dn in directives) {
			$el.find("["+directives[dn][0]+"]").each(function() {
				directives[dn][1]( $(this), scope );
			});
		}
		
		// Apply startup values
		scope.__refresh();
	};
	
	var startup = function() {
		var baseScope = new Scope();
		baseScope.opengularVersion = "opengular.js/1.X";
		
		$("[og-controller]").each(function() {
				
			var controllerName = $(this).attr('og-controller');
				
			// Create a new scope
			var s = new Scope();
			s.__parent = baseScope;
			$(this).data('og-scope', s);
			
			// Run the controller function to set up any handlers and default data for it
			controllers[controllerName](s);
			
			initDiv( $(this), s );
		});
	};
		
	$(function() {
		if ($("html").attr('og-autostart') === "") {
			startup();
		}
	});
	
	// -[PUBLIC API]-----------------
	
	return {
		'registerController': registerController,
		'registerDirective': registerDirective,
		'startup': startup
	};
	
})();
